# -*- coding: utf-8 -*-
import logging
import os
import time

from qiniu import compat
from qiniu import utils

UC_HOST = 'https://uc.qbox.me'  # 获取空间信息Host


class Region(object):
    """七牛上传区域类
    该类主要内容上传区域地址。
    """

    def __init__(
            self,
            up_host=None,
            up_host_backup=None,
            io_host=None,
            host_cache=None,
            home_dir=None,
            scheme="http",
            rs_host=None,
            rsf_host=None,
            api_host=None):
        """初始化Zone类"""
        if host_cache is None:
            host_cache = {}
        self.up_host = up_host
        self.up_host_backup = up_host_backup
        self.io_host = io_host
        self.rs_host = rs_host
        self.rsf_host = rsf_host
        self.api_host = api_host
        self.home_dir = home_dir
        self.host_cache = host_cache
        self.scheme = scheme

    def get_up_host_by_token(self, up_token, home_dir):
        ak, bucket = self.unmarshal_up_token(up_token)
        if home_dir is None:
            home_dir = os.getcwd()
        up_hosts = self.get_up_host(ak, bucket, home_dir)
        return up_hosts[0]

    def get_up_host_backup_by_token(self, up_token, home_dir):
        ak, bucket = self.unmarshal_up_token(up_token)
        if home_dir is None:
            home_dir = os.getcwd()
        up_hosts = self.get_up_host(ak, bucket, home_dir)
        if len(up_hosts) <= 1:
            up_host = up_hosts[0]
        else:
            up_host = up_hosts[1]
        return up_host

    def get_io_host(self, ak, bucket, home_dir=None):
        if self.io_host:
            return self.io_host
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'ioHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        io_hosts = bucket_hosts['ioHosts']
        return io_hosts[0]

    def get_rs_host(self, ak, bucket, home_dir=None):
        from .config import get_default, is_customized_default
        if self.rs_host:
            return self.rs_host
        if is_customized_default('default_rs_host'):
            return get_default('default_rs_host')
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'rsHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        rs_hosts = bucket_hosts['rsHosts']
        return rs_hosts[0]

    def get_rsf_host(self, ak, bucket, home_dir=None):
        from .config import get_default, is_customized_default
        if self.rsf_host:
            return self.rsf_host
        if is_customized_default('default_rsf_host'):
            return get_default('default_rsf_host')
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'rsfHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        rsf_hosts = bucket_hosts['rsfHosts']
        return rsf_hosts[0]

    def get_api_host(self, ak, bucket, home_dir=None):
        from .config import get_default, is_customized_default
        if self.api_host:
            return self.api_host
        if is_customized_default('default_api_host'):
            return get_default('default_api_host')
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'apiHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        api_hosts = bucket_hosts['apiHosts']
        return api_hosts[0]

    def get_up_host(self, ak, bucket, home_dir):
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'upHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        up_hosts = bucket_hosts['upHosts']
        return up_hosts

    def unmarshal_up_token(self, up_token):
        token = up_token.split(':')
        if (len(token) != 3):
            raise ValueError('invalid up_token')

        ak = token[0]
        policy = compat.json.loads(
            compat.s(
                utils.urlsafe_base64_decode(
                    token[2])))

        scope = policy["scope"]
        bucket = scope
        if (':' in scope):
            bucket = scope.split(':')[0]

        return ak, bucket

    def get_bucket_hosts(self, ak, bucket, home_dir, force=False):
        key = self.scheme + ":" + ak + ":" + bucket

        bucket_hosts = self.get_bucket_hosts_to_cache(key, home_dir)
        if not force and len(bucket_hosts) > 0:
            return bucket_hosts

        hosts = compat.json.loads(self.bucket_hosts(ak, bucket)).get('hosts', [])

        if type(hosts) is not list or len(hosts) == 0:
            raise KeyError("Please check your BUCKET_NAME! Server hosts not correct! The hosts is %s" % hosts)

        region = hosts[0]

        default_ttl = 24 * 3600  # 1 day
        region['ttl'] = region.get('ttl', default_ttl)

        bucket_hosts = {
            'upHosts': [
                '{0}://{1}'.format(self.scheme, domain)
                for domain in region.get('up', {}).get('domains', [])
            ],
            'ioHosts': [
                '{0}://{1}'.format(self.scheme, domain)
                for domain in region.get('io', {}).get('domains', [])
            ],
            'rsHosts': [
                '{0}://{1}'.format(self.scheme, domain)
                for domain in region.get('rs', {}).get('domains', [])
            ],
            'rsfHosts': [
                '{0}://{1}'.format(self.scheme, domain)
                for domain in region.get('rsf', {}).get('domains', [])
            ],
            'apiHosts': [
                '{0}://{1}'.format(self.scheme, domain)
                for domain in region.get('api', {}).get('domains', [])
            ],
            'deadline': int(time.time()) + region['ttl']
        }
        home_dir = ""
        self.set_bucket_hosts_to_cache(key, bucket_hosts, home_dir)
        return bucket_hosts

    def get_bucket_hosts_to_cache(self, key, home_dir):
        ret = {}
        if len(self.host_cache) == 0:
            self.host_cache_from_file(home_dir)

        if key not in self.host_cache:
            return ret

        if self.host_cache[key]['deadline'] > time.time():
            ret = self.host_cache[key]

        return ret

    def set_bucket_hosts_to_cache(self, key, val, home_dir):
        self.host_cache[key] = val
        self.host_cache_to_file(home_dir)
        return

    def host_cache_from_file(self, home_dir):
        if home_dir is not None:
            self.home_dir = home_dir
        path = self.host_cache_file_path()
        if not os.path.isfile(path):
            return None
        with open(path, 'r') as f:
            try:
                bucket_hosts = compat.json.load(f)
                self.host_cache = bucket_hosts
            except Exception as e:
                logging.error(e)
        f.close()
        return

    def host_cache_file_path(self):
        return os.path.join(self.home_dir, ".qiniu_pythonsdk_hostscache.json")

    def host_cache_to_file(self, home_dir):
        path = self.host_cache_file_path()
        with open(path, 'w') as f:
            compat.json.dump(self.host_cache, f)
        f.close()

    def bucket_hosts(self, ak, bucket):
        from .config import get_default, is_customized_default
        from .http import qn_http_client
        from .http.middleware import RetryDomainsMiddleware
        uc_host = UC_HOST
        if is_customized_default('default_uc_host'):
            uc_host = get_default('default_uc_host')
        uc_backup_hosts = get_default('default_query_region_backup_hosts')
        uc_backup_retry_times = get_default('default_backup_hosts_retry_times')
        url = "{0}/v4/query?ak={1}&bucket={2}".format(uc_host, ak, bucket)

        ret, _resp = qn_http_client.get(
            url,
            middlewares=[
                RetryDomainsMiddleware(
                    backup_domains=uc_backup_hosts,
                    max_retry_times=uc_backup_retry_times,
                )
            ]
        )
        data = compat.json.dumps(ret, separators=(',', ':'))
        return data



import abc

import qiniu.config as config

# type import
from qiniu.auth import Auth # noqa


class UploaderBase(object):
    """
    Attributes
    ----------
    bucket_name: str
    auth: Auth
    regions: list[Region]
    """
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        bucket_name,
        **kwargs
    ):
        """
        Parameters
        ----------
        bucket_name: str
            The name of bucket which you want to upload to.
        auth: Auth
            The instance of Auth to sign requests.
        regions: list[Region], default=[]
            The regions of bucket. It will be queried if not specified.
        kwargs
            The others arguments may be used by subclass.
        """
        self.bucket_name = bucket_name

        # change the default when implements AuthProvider
        self.auth = kwargs.get('auth', None)

        regions = kwargs.get('regions', [])
        # remove the check when implement RegionsProvider
        # if not regions:
        #     raise TypeError('You must provide the regions')
        self.regions = regions

        hosts_cache_dir = kwargs.get('hosts_cache_dir', None)
        self.hosts_cache_dir = hosts_cache_dir

    def get_up_token(self, **kwargs):
        """
        Generate up token

        Parameters
        ----------
        bucket_name: str
        key: str
        expired: int
        policy: dict
        strict_policy: bool

        Returns
        -------
        str
        """
        if not self.auth:
            raise ValueError('can not get up_token by auth not provided')

        bucket_name = kwargs.get('bucket_name', self.bucket_name)

        kwargs_for_up_token = {
            k: kwargs[k]
            for k in [
                'key', 'expires', 'policy', 'strict_policy'
            ] if k in kwargs
        }
        up_token = self.auth.upload_token(
            bucket=bucket_name,
            **kwargs_for_up_token
        )
        return up_token

    def _get_regions(self):
        if self.regions:
            return self.regions

        default_region = config.get_default('default_zone')
        if default_region:
            self.regions = [default_region]

        return self.regions

    def _get_up_hosts(self, access_key=None):
        """
        This will be deprecated when implement regions and endpoints

        Returns
        -------
        list[str]
        """
        if not self.auth and not access_key:
            raise ValueError('Must provide access_key if auth is unavailable.')
        if not access_key:
            access_key = self.auth.get_access_key()

        regions = self._get_regions()

        if not regions:
            raise ValueError('No region available.')

        if regions[0].up_host and regions[0].up_host_backup:
            return [
                regions[0].up_host,
                regions[0].up_host_backup
            ]

        # this is correct, it does return hosts. bad function name by legacy
        return regions[0].get_up_host(
            ak=access_key,
            bucket=self.bucket_name,
            home_dir=self.hosts_cache_dir
        )

    @abc.abstractmethod
    def upload(
        self,
        key,
        file_path,
        data,
        data_size,
        modify_time,

        part_size,
        mime_type,
        metadata,
        file_name,
        custom_vars,
        **kwargs
    ):
        """
        Upload method

        Parameters
        ----------
        key: str
        file_path: str
        data: IOBase
        data_size: int
        modify_time: int

        part_size: int
        mime_type: str
        metadata: dict
        file_name: str
        custom_vars: dict
        kwargs: dict

        Returns
        -------
        ret: dict
            The parsed response body
        info
            The response
        """



from collections import namedtuple
from concurrent import futures
from io import BytesIO
from os import path
from threading import Lock
from time import time

from qiniu.compat import is_seekable
from qiniu.auth import Auth
from qiniu.http import qn_http_client, ResponseInfo
from qiniu.utils import b, io_md5, urlsafe_base64_encode
from qiniu.compat import json

from qiniu.services.storage.uploaders.abc import ResumeUploaderBase
from qiniu.services.storage.uploaders.io_chunked import IOChunked


class ResumeUploaderV2(ResumeUploaderBase):
    def _recover_from_record(
        self,
        file_name,
        key,
        context
    ):
        """

        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV2Context

        Returns
        -------
        _ResumeUploadV2Context
        """
        if not isinstance(context, _ResumeUploadV2Context):
            raise TypeError('context must be an instance of _ResumeUploadV2Context')

        if (
            not self.upload_progress_recorder or
            not (file_name or key)
        ):
            return context

        record = self.upload_progress_recorder.get_upload_record(
            file_name,
            key
        )

        if not record:
            return context

        record_up_hosts = record.get('up_hosts', [])
        record_upload_id = record.get('upload_id', '')
        record_expired_at = record.get('expired_at', 0)
        record_part_size = record.get('part_size', None)
        record_modify_time = record.get('modify_time', 0)
        record_etags = record.get('etags', [])

        # compact with old sdk(<= v7.11.1)
        if not record_up_hosts or not record_part_size:
            return context

        if (
            not record_upload_id or
            record_modify_time != context.modify_time or
            record_expired_at > time()
        ):
            return context

        return context._replace(
            up_hosts=record_up_hosts,
            upload_id=record_upload_id,
            expired_at=record_expired_at,
            part_size=record_part_size,
            parts=[
                _ResumeUploadV2Part(
                    part_no=p['partNumber'],
                    etag=p['etag']
                )
                for p in record_etags
            ],
            resumed=True
        )

    def _set_to_record(self, file_name, key, context):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV2Context

        """
        if not self.upload_progress_recorder or not any([file_name, key]):
            return

        record_data = {
            'up_hosts': context.up_hosts,
            'upload_id': context.upload_id,
            'expired_at': context.expired_at,
            'part_size': context.part_size,
            'modify_time': context.modify_time,
            'etags': [
                {
                    'etag': part.etag,
                    'part_no': part.part_no
                }
                for part in context.parts
            ]
        }
        self.upload_progress_recorder.set_upload_record(
            file_name,
            key,
            data=record_data
        )

    def _try_delete_record(
        self,
        file_name,
        key,
        context,
        resp
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV2Context
        resp: ResponseInfo
        """
        if not self.upload_progress_recorder or not any([file_name, key]):
            return
        if resp and context and not any([
            resp.ok(),
            resp.status_code == 612 and context.resumed
        ]):
            return
        self.upload_progress_recorder.delete_upload_record(file_name, key)

    def _progress_handler(
        self,
        file_name,
        key,
        context,
        uploaded_size,
        total_size
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV2Context
        uploaded_size: int
        total_size: int
        """
        self._set_to_record(file_name, key, context)
        if callable(self.progress_handler):
            self.progress_handler(uploaded_size, total_size)

    def initial_parts(
        self,
        up_token,
        key,
        file_path=None,
        data=None,
        data_size=None,
        modify_time=None,
        part_size=None,
        **kwargs
    ):
        """

        Parameters
        ----------
        up_token: str
        key: str
        file_path: str
        data: IOBase
        data_size: int
        modify_time: int
        part_size: int
        kwargs

        Returns
        -------
        ret: _ResumeUploadV2Context
        resp: ResponseInfo
        """
        # -- check and initial arguments
        # must provide file_path or data
        if not file_path and not data:
            raise TypeError('Must provide one of file_path or data.')
        if file_path and data:
            raise TypeError('Must provide only one of file_path or data.')

        # data must has length
        if not file_path and not data_size:
            raise TypeError('Must provide size if use data.')

        if not modify_time:
            if file_path:
                modify_time = int(path.getmtime(file_path))
            else:
                modify_time = int(time())

        if not part_size:
            part_size = self.part_size

        # -- initial context
        context = _ResumeUploadV2Context(
            up_hosts=[],
            upload_id='',
            expired_at=0,
            part_size=part_size,
            parts=[],
            modify_time=modify_time,
            resumed=False
        )

        # try to recover from record
        file_name = path.basename(file_path) if file_path else None
        context = self._recover_from_record(
            file_name,
            key,
            context
        )

        if (
            context.up_hosts and
            context.upload_id and
            context.expired_at
        ):
            return context, None

        # -- get a new upload id
        access_key, _, _ = Auth.up_token_decode(up_token)
        if not context.up_hosts:
            context.up_hosts.extend(self._get_up_hosts(access_key))

        bucket_name = Auth.get_bucket_name(up_token)

        resp = None
        for up_host in context.up_hosts:
            url = self.__get_url_for_upload(
                up_host,
                bucket_name,
                key
            )
            ret, resp = qn_http_client.post(
                url=url,
                data='',
                files=None,
                headers={
                    'Authorization': 'UpToken {}'.format(up_token)
                }
            )
            if not resp.ok() and not resp.need_retry():
                break
            if resp.ok() and ret:
                context = context._replace(
                    upload_id=ret.get('uploadId', ''),
                    expired_at=ret.get('expireAt', 0)
                )
                break

        return context, resp

    def upload_parts(
        self,
        up_token,
        data,
        data_size,
        context,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        data
        data_size: int
        context: _ResumeUploadV2Context
        kwargs
            key, file_name

        Returns
        -------
        part: _ResumeUploadV2Part
        resp: ResponseInfo

        """
        # initial arguments
        chunk_list = self.gen_chunk_list(
            size=data_size,
            chunk_size=context.part_size,
            uploaded_chunk_no_list=[
                p.part_no for p in context.parts
            ]
        )
        up_hosts = list(context.up_hosts)
        file_name = kwargs.get('file_name', None)
        key = kwargs.get('key', None)

        # initial upload state
        part, resp = None, None
        uploaded_size = 0
        lock = Lock()

        if not self.concurrent_executor:
            # upload sequentially
            for chunk in chunk_list:
                part, resp = self.__upload_part(
                    data=data,
                    chunk_info=chunk,
                    up_hosts=up_hosts,
                    up_token=up_token,
                    upload_id=context.upload_id,
                    key=key,
                    lock=lock
                )
                if not resp.ok():
                    return None, resp
                context.parts.append(part)
                uploaded_size += chunk.chunk_size
                self._progress_handler(
                    file_name=file_name,
                    key=key,
                    context=context,
                    uploaded_size=uploaded_size,
                    total_size=data_size
                )
        else:
            # upload concurrently
            future_chunk_dict = {}
            for chunk in chunk_list:
                ftr = self.concurrent_executor.submit(
                    self.__upload_part,
                    data=data,
                    chunk_info=chunk,
                    up_hosts=up_hosts,
                    up_token=up_token,
                    upload_id=context.upload_id,
                    key=key,
                    lock=lock
                )
                future_chunk_dict[ftr] = chunk

            first_failed_resp = None
            for ftr in futures.as_completed(future_chunk_dict):
                if ftr.cancelled():
                    continue
                elif ftr.exception():
                    if first_failed_resp:
                        continue
                    first_failed_resp = ResponseInfo(None, ftr.exception())
                    for not_done in filter(lambda f: not f.done(), future_chunk_dict):
                        not_done.cancel()
                else:
                    part, resp = ftr.result()
                    if not part:
                        if not first_failed_resp:
                            first_failed_resp = resp
                            for not_done in filter(lambda f: not f.done(), future_chunk_dict):
                                not_done.cancel()
                    else:
                        context.parts.append(part)
                        uploaded_size += future_chunk_dict[ftr].chunk_size
                        self._progress_handler(
                            file_name=file_name,
                            key=key,
                            context=context,
                            uploaded_size=uploaded_size,
                            total_size=data_size
                        )
            if first_failed_resp:
                return None, first_failed_resp

        return part, resp

    def complete_parts(
        self,
        up_token,
        data_size,
        context,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        data_size: int
        context: _ResumeUploadV2Context
        kwargs
            key, file_name, params, metadata
        Returns
        -------
            ret: dict
            resp: ResponseInfo
        """
        key = kwargs.get('key', None)
        file_name = kwargs.get('file_name', None)
        mime_type = kwargs.get('mime_type', None)
        params = kwargs.get('params', None)
        metadata = kwargs.get('metadata', None)

        # sort contexts
        sorted_parts = sorted(context.parts, key=lambda part: part.part_no)

        bucket_name = Auth.get_bucket_name(up_token)

        ret, resp = None, None
        for up_host in context.up_hosts:
            url = self.__get_url_for_upload(
                up_host,
                bucket_name,
                key,
                upload_id=context.upload_id
            )
            data = {
                'parts': [
                    {
                        'etag': p.etag,
                        'partNumber': p.part_no
                    }
                    for p in sorted_parts
                ],
                'fname': file_name,
                'mimeType': mime_type,
                'customVars': params,
                'metadata': metadata
            }
            ret, resp = qn_http_client.post(
                url=url,
                data=json.dumps(data),
                files=None,
                headers={
                    'Content-Type': 'application/json',
                    'Authorization': 'UpToken {}'.format(up_token)
                }
            )
            if resp.ok() or not resp.need_retry():
                break
        self._try_delete_record(
            file_name,
            key,
            context,
            resp
        )
        return ret, resp

    def upload(
        self,
        key,
        file_path=None,
        data=None,
        data_size=None,

        part_size=None,
        modify_time=None,
        mime_type=None,
        metadata=None,
        file_name=None,
        custom_vars=None,
        **kwargs
    ):
        """
        Parameters
        ----------
        key: str
        file_path: str
        data: IOBase
        data_size: int
        part_size: int
        modify_time: int
        mime_type: str
        metadata: dict
        file_name: str
        custom_vars: dict
        kwargs
            up_token
            bucket_name, expires, policy, strict_policy for generate `up_token`

        Returns
        -------

        """
        # up_token
        up_token = kwargs.get('up_token', None)
        if not up_token:
            up_token = self.get_up_token(**kwargs)
        if not file_name and file_path:
            file_name = path.basename(file_path)

        # initial_parts
        context, resp = self.initial_parts(
            up_token,
            key,
            file_path=file_path,
            data=data,
            data_size=data_size,
            modify_time=modify_time,
            part_size=part_size
        )

        if (
            not context.up_hosts or
            not context.upload_id or
            not context.expired_at
        ):
            return None, resp

        # upload parts
        try:
            if file_path:
                data_size = path.getsize(file_path)
                data = open(file_path, 'rb')
            elif isinstance(data, bytes):
                data_size = len(data)
                data = BytesIO(data)
            elif isinstance(data, str):
                data_size = len(data)
                data = BytesIO(b(data))
            ret, resp = self.upload_parts(
                up_token=up_token,
                data=data,
                data_size=data_size,
                context=context,

                key=key,
                file_name=file_name
            )
        finally:
            if file_path:
                data.close()

        if resp and not resp.ok():
            if resp.status_code == 612 and context.resumed:
                return self.upload(
                    key,
                    file_path=file_path,
                    data=data,
                    data_size=data_size,
                    modify_time=modify_time,

                    mime_type=mime_type,
                    metadata=metadata,
                    file_name=file_name,
                    custom_vars=custom_vars,
                    **kwargs
                )
            return ret, resp

        # complete parts
        ret, resp = self.complete_parts(
            up_token=up_token,
            data_size=data_size,
            context=context,

            key=key,
            mime_type=mime_type,
            file_name=file_name,
            params=custom_vars,
            metadata=metadata
        )

        # retry if expired. the record file will be deleted by complete_parts
        if resp.status_code == 612 and context.resumed:
            return self.upload(
                key,
                file_path=file_path,
                data=data,
                data_size=data_size,
                modify_time=modify_time,

                mime_type=mime_type,
                metadata=metadata,
                file_name=file_name,
                custom_vars=custom_vars,
                **kwargs
            )

        return ret, resp

    def __get_url_for_upload(
        self,
        up_host,
        bucket_name,
        key,
        upload_id=None,
        part_no=None,
    ):
        """
        Parameters
        ----------
        up_host: str
        bucket_name: str
        key: str
        upload_id: str
        part_no: int

        Returns
        -------
        str
        """
        if not bucket_name:
            bucket_name = self.bucket_name

        object_entry = '~'
        if key:
            object_entry = urlsafe_base64_encode(key)

        url_segs = [
            up_host,
            'buckets', bucket_name,
            'objects', object_entry,
            'uploads',
        ]

        if upload_id:
            url_segs.append(upload_id)

        if part_no:
            url_segs.append(str(part_no))

        return '/'.join(url_segs)

    def __upload_part(
        # resort arguments
        self,
        data,
        chunk_info,
        up_hosts,
        up_token,
        upload_id,
        key,
        lock
    ):
        """
        Parameters
        ----------
        data: IOBase
        chunk_info: ChunkInfo
        up_hosts: list[str]
        up_token: str
        upload_id: str
        key: str
        lock: Lock

        Returns
        -------
        part: _ResumeUploadV2Part
        resp: ResponseInfo
        """
        if not up_hosts:
            raise ValueError('Must provide on up host at least')

        bucket_name = Auth.get_bucket_name(up_token)
        if not bucket_name:
            bucket_name = self.bucket_name

        chunked_data = IOChunked(
            base_io=data,
            chunk_offset=chunk_info.chunk_offset,
            chunk_size=chunk_info.chunk_size,
            lock=lock
        )
        chunk_md5 = io_md5(chunked_data)
        chunked_data.seek(0)
        part, resp = None, None
        for up_host in up_hosts:
            url = self.__get_url_for_upload(
                up_host,
                bucket_name,
                key,
                upload_id=upload_id,
                part_no=chunk_info.chunk_no
            )
            ret, resp = qn_http_client.put(
                url=url,
                data=chunked_data,
                files=None,
                headers={
                    'Content-Type': 'application/octet-stream',
                    'Content-MD5': chunk_md5,
                    'Authorization': 'UpToken {}'.format(up_token)
                }
            )
            if resp.ok() and ret:
                part = _ResumeUploadV2Part(
                    part_no=chunk_info.chunk_no,
                    etag=ret.get('etag', '')
                )
                return part, resp
            if (
                not is_seekable(chunked_data) or
                not resp.need_retry()
            ):
                return part, resp
            chunked_data.seek(0)
        return part, resp


# use dataclass instead namedtuple if min version of python update to 3.7
_ResumeUploadV2Part = namedtuple(
    'ResumeUploadV2Part',
    [
        'part_no',
        'etag'
    ]
)

_ResumeUploadV2Context = namedtuple(
    'ResumeUploadV2Context',
    [
        'up_hosts',
        'upload_id',
        'expired_at',
        'part_size',
        'parts',
        'modify_time',
        'resumed'
    ]
)



# -*- coding: utf-8 -*-
import os

from qiniu.config import _BLOCK_SIZE, get_default

from qiniu.auth import Auth
from qiniu.utils import crc32, file_crc32, rfc_from_timestamp

from qiniu.services.storage.uploaders import FormUploader, ResumeUploaderV1, ResumeUploaderV2
from qiniu.services.storage.upload_progress_recorder import UploadProgressRecorder

# for compact to old sdk
from qiniu.services.storage.legacy import _Resume # noqa


def put_data(
    up_token, key, data, params=None, mime_type='application/octet-stream', check_crc=False, progress_handler=None,
    fname=None, hostscache_dir=None, metadata=None
):
    """上传二进制流到七牛

    Args:
        up_token:         上传凭证
        key:              上传文件名
        data:             上传二进制流
        params:           自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
        mime_type:        上传数据的mimeType
        check_crc:        是否校验crc32
        progress_handler: 上传进度
        hostscache_dir:   host请求 缓存文件保存位置
        metadata:         元数据

    Returns:
        一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
        一个ResponseInfo对象
    """
    final_data = b''
    if hasattr(data, 'read'):
        while True:
            tmp_data = data.read(_BLOCK_SIZE)
            if len(tmp_data) == 0:
                break
            else:
                final_data += tmp_data
    else:
        final_data = data

    crc = crc32(final_data)
    return _form_put(
        up_token, key, final_data, params, mime_type,
        crc, hostscache_dir, progress_handler, fname, metadata=metadata
    )


def put_file(
    up_token, key, file_path, params=None,
    mime_type='application/octet-stream', check_crc=False,
    progress_handler=None, upload_progress_recorder=None, keep_last_modified=False, hostscache_dir=None,
    part_size=None, version=None, bucket_name=None, metadata=None
):
    """上传文件到七牛

    Args:
        up_token:                 上传凭证
        key:                      上传文件名
        file_path:                上传文件的路径
        params:                   自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
        mime_type:                上传数据的mimeType
        check_crc:                是否校验crc32
        progress_handler:         上传进度
        upload_progress_recorder: 记录上传进度，用于断点续传
        hostscache_dir:           host请求 缓存文件保存位置
        version:                  分片上传版本 目前支持v1/v2版本 默认v1
        part_size:                分片上传v2必传字段 默认大小为4MB 分片大小范围为1 MB - 1 GB
        bucket_name:              分片上传v2字段 空间名称
        metadata:                 元数据信息

    Returns:
        一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
        一个ResponseInfo对象
    """
    ret = {}
    size = os.stat(file_path).st_size
    with open(file_path, 'rb') as input_stream:
        file_name = os.path.basename(file_path)
        modify_time = int(os.path.getmtime(file_path))
        if size > get_default('default_upload_threshold'):
            ret, info = put_stream(
                up_token, key, input_stream, file_name, size, hostscache_dir, params,
                mime_type, progress_handler,
                upload_progress_recorder=upload_progress_recorder,
                modify_time=modify_time, keep_last_modified=keep_last_modified,
                part_size=part_size, version=version, bucket_name=bucket_name, metadata=metadata
            )
        else:
            crc = file_crc32(file_path)
            ret, info = _form_put(
                up_token, key, input_stream, params, mime_type,
                crc, hostscache_dir, progress_handler, file_name,
                modify_time=modify_time, keep_last_modified=keep_last_modified, metadata=metadata
            )
    return ret, info


def _form_put(
    up_token,
    key,
    data,
    params,
    mime_type,
    crc,
    hostscache_dir=None,
    progress_handler=None,
    file_name=None,
    modify_time=None,
    keep_last_modified=False,
    metadata=None
):
    bucket_name = Auth.get_bucket_name(up_token)
    uploader = FormUploader(
        bucket_name,
        progress_handler=progress_handler,
        hosts_cache_dir=hostscache_dir
    )

    if modify_time and keep_last_modified:
        metadata['x-qn-meta-!Last-Modified'] = rfc_from_timestamp(modify_time)

    return uploader.upload(
        key=key,
        data=data,
        data_size=None,
        file_name=file_name,
        modify_time=modify_time,
        mime_type=mime_type,
        metadata=metadata,
        params=params,
        crc32_int=crc,
        up_token=up_token
    )


def put_stream(
    up_token,
    key,
    input_stream,
    file_name,
    data_size,
    hostscache_dir=None,
    params=None,
    mime_type=None,
    progress_handler=None,
    upload_progress_recorder=None,
    modify_time=None,
    keep_last_modified=False,
    part_size=None,
    version='v1',
    bucket_name=None,
    metadata=None
):
    if not bucket_name:
        bucket_name = Auth.get_bucket_name(up_token)
    if not upload_progress_recorder:
        upload_progress_recorder = UploadProgressRecorder()
    if not version:
        version = 'v1'
    if not part_size:
        part_size = 4 * (1024 * 1024)

    if version == 'v1':
        uploader = ResumeUploaderV1(
            bucket_name,
            progress_handler=progress_handler,
            upload_progress_recorder=upload_progress_recorder,
            hosts_cache_dir=hostscache_dir
        )
        if modify_time and keep_last_modified:
            metadata['x-qn-meta-!Last-Modified'] = rfc_from_timestamp(modify_time)
    elif version == 'v2':
        uploader = ResumeUploaderV2(
            bucket_name,
            progress_handler=progress_handler,
            upload_progress_recorder=upload_progress_recorder,
            part_size=part_size,
            hosts_cache_dir=hostscache_dir
        )
    else:
        raise ValueError('version only could be v1 or v2')
    return uploader.upload(
        key=key,
        data=input_stream,
        data_size=data_size,
        file_name=file_name,
        modify_time=modify_time,
        mime_type=mime_type,
        metadata=metadata,
        params=params,
        up_token=up_token
    )
